/*
    Elfproef- en betaalkenmerktester.

    Copyright 1999, R.Harmsen

    mei 2009   : Adsense-reclame toevoegd
    23 jan 2012: Sitemeter-code toegevoegd
     7 nov 2012: Sitemeter-code weer disabled
     5 Dec 2014: Added meta viewport
    10 mei 2021: Adsense-reclame verwijderd

    16 nov 2022: Fix a bug, which caused a buffer overflow,
                 and possibly corrupted malloc information.

    5 February 2023: Better CSS mechanism.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include "../cgitools/cgitools.h"

enum algor_enum
{
   ALGOR_ELFPROEF,
   ALGOR_BETAALKENMERK
};

void showform (char *script, char *input, enum algor_enum algor);
void htmlhead (void);
void htmlfoot (void);
void getparms (int len, char *input, enum algor_enum *p_algor);

int calculate (char *input, enum algor_enum algor);
int calc_and_show (
   char *betkenm,
   char *chkdig,
   enum algor_enum algor);

void digits_only (char *input);

int OutputSitemeter (FILE *fp);

int main (void)
{
   int len;
   #ifdef __MSDOS__
   // char *scriptname = "//127.0.0.1/cgi-bin/chkdigit.cgi";
   // char *scriptname = "//localhost/cgi-bin/chkdigit.cgi";
   #else
   // char *scriptname = "//utopia.knoware.nl/~rharmsen/cgi-bin/chkdigit.cgi";
   #endif
   char *scriptname = "../cgi-bin/chkdigit.cgi";

   char input[19+1] = "0123456789";
   enum algor_enum algor = ALGOR_ELFPROEF;

   if ((len = userinput()) == 0)
   {
      htmlhead();
      showform(scriptname, input, algor);
      htmlfoot();
   }
   else
   {
      getparms(len, input, &algor);
      htmlhead();
      showform(scriptname, input, algor);
      calculate(input, algor);
      htmlfoot();
   }
   return 0;
}


void showform (char *script, char *input, enum algor_enum algor)
{
   printf(
      "<FORM ACTION=\"%s\" METHOD=POST>\n"
      "<h1>Controlecijfertester.</h1>\n"
      "<!--"
      "<H3><A TARGET = _top HREF="
      "\"../cgi-bin/donation.cgi?lang=nl&grains=3&url=../chkdigit/chkdigit.htm\">"
      "Donatie</A></H3>\n"
      "-->"
      "<p>Controleert een ingegeven bankrekeningnummer of betaalkenmerk\n"
      "volgens de elfproef of op controlecijfer van een betaalkenmerk,\n"
      "en laat zien hoe de berekening gegaan is.</p>"
      "\n"
      "<INPUT TYPE=Submit VALUE=\"Bereken\">\n"
      "<INPUT TYPE=Text NAME=Input SIZE=20 MAXLENGTH=19 VALUE=%s>"
      " Te testen nummer\n"
      "<br><br>\n"
      "<INPUT TYPE=Radio NAME=Algor VALUE=Elfproef %s>"
      " Elfproef bankrekeningnummer<br>\n"
      "<INPUT TYPE=Radio NAME=Algor VALUE=Betkenm %s>"
      " Controlecijfer betaalkenmerk\n"
      "</FORM>\n\n",
      script, input,
      algor == ALGOR_ELFPROEF      ? "Checked" : "",
      algor == ALGOR_BETAALKENMERK ? "Checked" : "");
}

void explkenmerk (void)
{
   printf(
      "\n<p>\n"
      "De informatie over de berekening van het controlecijfer\n"
      "van het betaalkenmerk was afkomstig van de ING-bank, uit\n"
      "een beschrijving van SWIFT MT-940. Maar in modernere\n"
      "downloadbare versies daarvan vind ik die beschrijving\n"
      "helaas niet meer terug.\n"
      "<br><br>\n"
      "Na een tip van iemand van het CWI zag ik in dat de factoren\n"
      "gewoon machten van twee modulo elf zijn. Uiteindelijk geeft dat\n"
      "natuurlijk hetzelfde effect, alleen de tussenresultaten blijven\n"
      "kleiner omdat de modulofunctie ook al steeds op elke factor wordt\n"
      "toegepast.\n"
      "<br>\n"
      "Om dat te illustreren zijn er nu drie kolommen, met in de derde\n"
      "de factoren in C-bitschuifnotatie.\n"
      "\n</p>\n"
      );
}

void htmlhead (void)
{
   printf(
      "Content-type: text/html\n\n"
      "<!doctype html>"
      "<html>\n"
      "<head>\n"
      "<title>Check digit calculation</title>\n"
      "<meta name=viewport content=\"width=device-width, initial-scale=1\">\n"
      "<link rel=stylesheet href=\"../css/cpyrport.css\">\n"
      "<link rel=stylesheet href=\"../cgi-bin/colschem.cgi\">\n"
      "<style>body {margin: 10px;}</style>\n"
      "</head>\n"
      "<body>\n"
      );

   fflush(stdout);
}

void htmlfoot (void)
{
   printf(
      "<p class=NCol>\n"
      "<a class=NCol href=\"../index/colexpnl.htm\">Kleuren:</a>\n"
      "<a class=NCol href=\"../cgi-bin/colcooky.cgi?NoWeirdColours=1\">Neutraal</a>\n"
      "<a class=NCol href=\"../cgi-bin/colcooky.cgi?NoWeirdColours=0\">Raar</a>\n"
      "<a class=NCol href=\"javascript:location.reload();\">Opnieuw laden</a>\n"
      "</p>\n"
      );

   printf(
      "<p class=pagelink>\n"
      "<a href=\"../index/math.htm\">Wiskundemenu</a>\n"
      "<a top href=\"../index/menu.htm\">Hoofdmenu</a>\n"
      "</p>"
      );

   printf("</body>\n</html>\n");
   fflush(stdout);
}

void getparms (int len, char *input, enum algor_enum *p_algor)
{
   char *eq1, *eq2;
   char *amp;
   char *buf = calloc(len, 1);

   if (!buf ||
      fread(buf, len, 1, stdin) != 1)
   {
      free(buf);
      return;
   }

   eq1 = strchr(buf, '=');
   if (!eq1)
   {
      free(buf);
      return;
   }

   eq1++;
   if ((amp = strchr(eq1, '&')) != NULL)
   {
      *amp = '\0';
   }
   strcpy(input, eq1);

   digits_only(input);

   if (amp)
      eq1 = amp + 1;

   eq2 = strchr(eq1, '=');
   if (!eq2)
   {
      free(buf);
      return;
   }

   eq2++;
   if (strncmp(eq2, "Betkenm", strlen("Betkenm")) == 0)
      *p_algor = ALGOR_BETAALKENMERK;
   else if (strncmp(eq2, "Elfproef", strlen("Elfproef")) == 0)
      *p_algor = ALGOR_ELFPROEF;

   if (strlen(input) > 10)
      *p_algor = ALGOR_BETAALKENMERK;

   free(buf);
}

/*
static int elfproef_gewicht[10] =
{
   10, 9, 8, 7, 6, 5, 4, 3, 2, 1,
};

Na een tip van Dik T. Winter <Dik.Winter@cwi.nl> werd duidelijk dat
dit steeds 2-machten modulo-11 zijn: 5 is 16%11, 10 is 32%11, enz.
Daarom worden ze nu niet uit de tabel gehaald, maar berekend.
static int betkenm_gewicht[16] =
{
   0, 10, 5, 8, 4, 2, 1, 6, 3, 7, 9, 10, 5, 8, 4, 2,
};
*/

/***********************************************************************
   Controleer betalingskenmerk (controlecijfer)

   Returns 1 if the number is correct (has the right check digit as
   first position), 0 if not.
   betkenm must include the check digit, and have a max length of 16+1
   (closing nullbyte included).

   chkdigit is a pointer, which if NULL will be ignored, but otherwise
   will contain the correct check digit afterwards (without a closing
   nullbyte, jyst the ASCII-digit).
 **********************************************************************/
int calculate (char *input, enum algor_enum algor)
{
   char chkdig;

   return calc_and_show(input, &chkdig, algor);
}

int calc_and_show (
   char *betkenm,
   char *chkdig,
   enum algor_enum algor)
{
   long sum1 = 0;
   long sum2 = 0;
   char *p;
   long chk1;
   long chk2;
   int corr;
   /* Internal buffer */
   char kenm[16+1];
   size_t l;
   long numdig = algor == ALGOR_BETAALKENMERK ? 16 : 10;

   printf("<pre><font size=-2>");

   sprintf(kenm, "%-16.16s", betkenm);
   if ((l = strspn(kenm, "0123456789")) > 0 && l < 16)
      kenm[l] = '\0';

   corr = numdig - strlen(kenm);
   for (p = algor == ALGOR_BETAALKENMERK ? kenm + 1 : kenm; *p; p++)
   {
      int dig = (*p - '0');
      int exp = numdig - (p - kenm + corr);
      long fac;
      long prd1;
      long prd2;

      if (algor == ALGOR_BETAALKENMERK)
      {
         fac = 1L << exp;
         prd1 = dig * (fac % 11);
         prd2 = dig * fac;

         printf("%d * %2ld = %2ld\t"   , dig, fac%11, prd1);
         printf("%d * %5ld = %6ld\t"   , dig, fac   , prd2);
         printf("%d * 1<<%-2d = %6ld\n", dig, exp   , prd2);
         sum1 += prd1;
         sum2 += prd2;
      }
      else
      {
         fac  = exp;
         prd2 = dig * fac;
         printf("%d * %2ld = %2ld\n", dig, fac, prd2);
         sum2 += prd2;
      }
   }
   if (algor == ALGOR_BETAALKENMERK)
   {
      printf("-----------+    ------------------+\n");
      printf("        %3ld                 %6ld\n", sum1, sum2);
   }
   else
   {
      printf("-----------+\n");
      printf("        %3ld\n", sum2);
   }

   printf("</font></pre>\n");

   if (algor == ALGOR_BETAALKENMERK)
   {
      chk2 = chk1 = 11 - sum1 % 11;
      if (chk1 == 10)
         chk2 = 1;
      else if (chk1 == 11)
         chk2 = 0;

      printf("Controlecijfer van \"%-16s\": 11 - (%6ld %% 11) = %2ld --> %ld\n",
         kenm, sum1, chk1, chk2);

      chk2 = chk1 = 11 - sum2 % 11;
      if (chk1 == 10)
         chk2 = 1;
      else if (chk1 == 11)
         chk2 = 0;

      printf("<br>");
      printf("Controlecijfer van \"%-16s\": 11 - (%6ld %% 11) = %2ld --> %ld\n",
         kenm, sum2, chk1, chk2);

   }
   else
   {
      printf("Rekeningnummer \"%-10s\": %ld %% 11 = %ld, dus ",
         kenm, sum2, sum2 % 11);
      printf("%s\n", sum2 % 11 == 0 ? "elfproef" : "niet elfproef");
   }
   if (chkdig)
      *chkdig = chk2 + '0';

   if (algor == ALGOR_BETAALKENMERK)
   {
      explkenmerk();
   }

   return chk2 + '0' == kenm[0];
}

void digits_only (char *input)
{
   char *p, *q;
   char *buf = calloc(strlen(input) + 1, 1);

   if (!buf)
      return;

   for (p = input, q = buf; *p; p++)
   {
      if (isdigit(*p))
         *q++ = *p;
   }
   *q = '\0';
   strcpy(input, buf);
   free(buf);

   return;
}

#if 0
// 7 Nov 2012: Disabled, to avoid any problems with the EU Cookie Laws
static char insert_code[] =
{
   "<script type=\"text/javascript\" src=\"http://sm9.sitemeter.com/js/counter.js?site=sm9rudhar\">\n"
   "</script>\n"
   "<noscript>\n"
   "<a href=\"http://sm9.sitemeter.com/stats.asp?site=sm9rudhar\" target=\"_top\">\n"
   "<img src=\"http://sm9.sitemeter.com/meter.asp?site=sm9rudhar\" alt=\"Site Meter\" border=\"0\"/></a>\n"
   "</noscript>\n"
   "<!-- Copyright (c)2009 Site Meter -->\n"
};
#endif

int OutputSitemeter (FILE *fp)
{
// 7 Nov 2012: Disabled, to avoid any problems with the EU Cookie Laws
#if 0
   fwrite(insert_code, (sizeof insert_code) - 1, 1, fp);
#endif

   return 0;
}

